import time
import argparse
import itertools
import ismrmrd
import pathlib

import numpy as np
import matplotlib.pyplot as plt

from pynufft import NUFFT
from ugad4fun.utils.dummy_conn import DummyConn


def accumulate_acquisitions(acquisitions, header):
    accumulated_acquisitions = []
    matrix_size = header.encoding[0].encodedSpace.matrixSize
    print("matrix_size.x=" + str(matrix_size.x))
    print("matrix_size.y=" + str(matrix_size.y))
    print("matrix_size.z=" + str(matrix_size.z))

    def assemble_buffer(acqs):
        print(f"Assembling buffer from {len(acqs)} acquisitions.")

        number_of_channels, number_of_samples = acqs[0].data.shape
        trajectory_dimensions = acqs[0].trajectory_dimensions
        print("number_of_channels=" + str(number_of_channels))
        print("number_of_samples=" + str(number_of_samples))
        print("trajectory_dimensions=" + str(trajectory_dimensions))

        raw_buffer = np.zeros(
            (matrix_size.x,
             matrix_size.y,
             matrix_size.z,
             number_of_channels,
             ),
            dtype=np.complex64
        )

        traj_buffer = np.zeros(
            (matrix_size.x,
             matrix_size.y,
             matrix_size.z,
             trajectory_dimensions
             ),
            dtype=np.float32
        )

        for acq in acqs:
            data = acq.data
            traj = acq.traj
            
            raw_buffer[:, acq.idx.kspace_encode_step_1, acq.idx.kspace_encode_step_2, :] = acq.data.transpose(1, 0)
            
            # pynufft needs traj to be in the range of [-pi, pi]
            # traj = (traj - traj.min()) / (traj.max() - traj.min()) * 2 * np.pi - np.pi
            traj_buffer[:, acq.idx.kspace_encode_step_1, acq.idx.kspace_encode_step_2, :] = traj
        return raw_buffer, traj_buffer

    for acquisition in acquisitions:
        accumulated_acquisitions.append(acquisition)
        if acquisition.is_flag_set(ismrmrd.ACQ_LAST_IN_SLICE):
            yield acquisition, assemble_buffer(accumulated_acquisitions)
            accumulated_acquisitions = []


def reconstruct_images(buffers, header):

    indices = itertools.count(start=1)
    field_of_view = header.encoding[0].reconSpace.fieldOfView_mm
    kspace_size = header.encoding[0].encodedSpace.matrixSize
    matrix = header.encoding[0].reconSpace.matrixSize

    def combine_channels(img):
        # The buffer contains complex images, one for each channel. We combine these into a single image
        # through a sum of squares along the channels (axis 2).

        # return np.sqrt(np.sum(np.square(np.abs(img)), axis=2))
        combined_complex_img = np.zeros_like(img[:, :, 0])
        for x in range(img.shape[0]):
            for y in range(img.shape[1]):
                mag = 0
                phase = 0
                for ch in range(img.shape[2]):
                    img_data = img[x, y, ch]
                    mag_tmp = img_data.real ** 2 + img_data.imag ** 2
                    phase += mag_tmp * np.angle(img_data)
                    mag += mag_tmp
                combined_complex_img[x, y] = np.sqrt(mag) * np.exp(complex(0, phase / mag))
        return combined_complex_img
        
    def reconstruct_image(data):
        # Reconstruction is a pynufft fft in this case.
        kspace_data = data[0].squeeze()
        traj = data[1].squeeze()    
        print('Trajectory shape: ', traj.shape)
        print('K space data shape: ', kspace_data.shape)
        kspace_data_reshape = kspace_data.reshape(-1, kspace_data.shape[2])

        traj_reshape = traj.reshape(-1, traj.shape[2])

        # image size
        Nd = (matrix.x, matrix.y)
        print('setting image dimension Nd...', Nd)
        # k-space size
        Kd = (kspace_size.x, kspace_size.y)
        print('setting spectrum dimension Kd...', Kd)
        # interpolation size
        Jd = (6, 6)
        print('setting interpolation size Jd...', Jd)

        NufftObj = NUFFT()
        NufftObj.plan(traj_reshape, Nd, Kd, Jd)
        img = np.empty((Nd[0], Nd[1], kspace_data_reshape.shape[1]), dtype=np.complex128)
        for index, current_raw_data in enumerate(kspace_data_reshape.transpose(1, 0)):
            img[:, :, index] = NufftObj.solve(current_raw_data, solver='cg', maxiter=25)
        image_final = combine_channels(img)
        print('reconstruct images finished!')
        return image_final

    for reference, data in buffers:
        print("image_index = " + str(indices))
        yield ismrmrd.image.Image.from_array(
            reconstruct_image(data),
            acquisition=reference,
            image_index=next(indices),
            image_type=ismrmrd.IMTYPE_MAGNITUDE,
            # data_type=ismrmrd.DATATYPE_FLOAT,
            data_type=ismrmrd.DATATYPE_CXFLOAT,
            field_of_view=(field_of_view.x, field_of_view.y, field_of_view.z),
            transpose=True, # this will transpose the image data
        )
        
def recon(conn):
    print("Python reconstruction running - reconstructing images from acquisitions.")
    conn.filter(ismrmrd.acquisition.Acquisition)
    acquisitions = iter(conn)
    buffers = accumulate_acquisitions(acquisitions, conn.header)
    images = reconstruct_images(buffers, conn.header)

    for image in images:
        print("Sending image back to client.")
        print(f"image.matrix_size = '{image.matrix_size}'")
        conn.send(image)
    
    print("radial image recon done.")
    


if __name__ == '__main__':

    help_msg = "Radial recon using dummy connection"

    parser = argparse.ArgumentParser(description=help_msg)
    parser.add_argument('input_path', type=str,
                        help='the input path of an ismrmrd data')
    parser.add_argument('-o', '--out_path', type=str,
                        help='the output path of recon')
    
    args = parser.parse_args()
    input_path = pathlib.Path(args.input_path).absolute()
    output_path = pathlib.Path(args.out_path).absolute() if args.out_path else None

    start_time = time.time()
    dummy_conn = DummyConn(input_path, output_path)
    recon(dummy_conn)
    dummy_conn.close()

    print(f'time costed {round(time.time() - start_time, 2)}s')
 



import faulthandler
import logging
import multiprocessing

# from PIL import Image as PILImage
import gadgetron

def do_work(conn):
    for index, acq in enumerate(conn):
        print(rf'{index}: data type: {type(acq)}')
        conn.send(acq)
        pass
    pass

# if __name__ == "__main__":
#     faulthandler.enable()
#     logging.basicConfig(level=logging.INFO, format="[%(levelname)s] %(message)s")

#     def spawn_process(*args):
#         child = multiprocessing.Process(target=do_work, args=args)
#         child.start()

#     while True:
#         gadgetron.external.listen(18000, spawn_process)


import gadgetron
import ismrmrd
from PIL import Image as PILImage
from matplotlib import pyplot as plt
from matplotlib.figure import Figure
import numpy as np
import logging
import multiprocessing
import typing
import faulthandler

def gen_image_to_send(header: ismrmrd.xsd.ismrmrdHeader,
                      acqs: typing.List[ismrmrd.Acquisition]=None,
                      connection: gadgetron.external.Connection=None):
    # target picture size, should get get from header in the future
    recon_matrix=header.encoding[0].reconSpace.matrixSize
    ic=1 # channel
    iz=1 # z size
    iy=recon_matrix.y # height
    ix=recon_matrix.x # width
    dpi=100 # data per inch
    figure=plt.gcf() # type: Figure
    figure.set_size_inches(ix/dpi,iy/dpi)

    print(rf'--------------------plot begin-----------------------')
    t = np.arange(0.0, 2.0, 0.01)
    s = np.sin(2 * np.pi * t)

    plt.plot(t, s)
    plt.title(r'$\alpha_i > \beta_i$', fontsize=20)
    plt.text(1, -0.6, r'$\sum_{i=0}^\infty x_i$', fontsize=20)
    plt.text(0.6, 0.6, r'$\mathcal{A}\mathrm{sin}(2 \omega t)$',
             fontsize=20)
    plt.xlabel('time (s)')
    plt.ylabel('volts (mV)')

    plt.savefig("result.png", dpi=dpi) # 3x100, 2x100
    print(rf'--------------------plot end-----------------------')

    h=ismrmrd.ImageHeader()# type: ismrmrd.ImageHeader
    #h.data_type = ismrmrd.DATATYPE_USHORT
    h.data_type = ismrmrd.DATATYPE_FLOAT
    h.image_type= ismrmrd.IMTYPE_MAGNITUDE
    h.channels=ic
    h.matrix_size=(ix,iy,iz)

    i=ismrmrd.Image(h)

    gray_scale_img=PILImage.open('result.png').convert('L') # gray scale image uint 8
    img_data=np.array(gray_scale_img)

    #TODO how to stand this code?
    #i.data[:] = img_data[:, ::-1].data[:]
    i.data[:]=img_data.data[:]
    return i

def do_work(connection:gadgetron.external.Connection):
    print('-----------------------begin------------------------')

    acqs=[] # type: typing.List[ismrmrd.Acquisition]

    for index, acq in enumerate(connection): # type: ismrmrd.Acquisition
        #do what you want todo?
        acqs.append(acq)
        pass

    print(rf'-------------------acq end-------------------')
    header=connection.header # type: ismrmrd.xsd.ismrmrdHeader
    image=gen_image_to_send(header, acqs, connection)

    print('  --begin-- try to send image')
    connection.send(image)
    print('  --end-- try to send image')

    print('-----------------------end------------------------')

if __name__ == "__main__":
    faulthandler.enable()
    logging.basicConfig(level=logging.INFO, format="[%(levelname)s] %(message)s")

    def spawn_process(*args):
        child = multiprocessing.Process(target=do_work, args=args)
        child.start()

    while True:
        gadgetron.external.listen(18000, spawn_process)



import ismrmrd

import numpy as np
import matplotlib.pyplot as plt


class DummyConn:
    def __init__(self, dataset_path, save_path=None):
        """A dummy connection for reconstructing radial rawdata locally

        Parameters
        ----------
        dataset_path : str | pathlib.Path
            ismrmrd dataset path (should be a .h5 file)
        save_path : str | pathlib.Path, optional
            recon result output path, by default None
            if None, plot images

        Raises
        ------
        NameError
            wrong input file format
        FileNotFoundError
            input file not exists
        """
        self.dataset_path = dataset_path
        self.save_path = save_path
        self.plot_mode = False if save_path else True
        self.filter_data = list()
        self.save_dataset = None

        # check dataset path
        if not str(dataset_path).endswith('.h5'):
            raise NameError("File format should be exactly .h5")
        if not self.dataset_path.exists():
            raise FileNotFoundError(rf"{self.dataset_path} Not Found")
        print(f'Loading {self.dataset_path}...')
        self.dataset = ismrmrd.Dataset(self.dataset_path)
        print(f'Extracting Header info...')
        self.header = ismrmrd.xsd.CreateFromDocument(self.dataset.read_xml_header())

    def filter(self, predicate, **kwargs):
        if not isinstance(predicate, type):
            raise ValueError(rf"predicate should be type ,get {predicate}")
        print(rf"Filtering {predicate}")
        if predicate is ismrmrd.Acquisition:
            print(f'{self.dataset.number_of_acquisitions()} acquisitions included')
            acquisition = []
            for acq_num in range(self.dataset.number_of_acquisitions()):
                acq = self.dataset.read_acquisition(acq_num)
                acquisition.append(acq)
            print(f'{len(acquisition)} raw data extracted')
            self.filter_data = iter(acquisition)
        elif predicate is ismrmrd.Waveform:
            print(f'{self.dataset.number_of_waveforms()} waveforms included')
            waveform = []
            for waveform_num in range(self.dataset.number_of_waveforms()):
                wave = self.dataset.read_waveform(waveform_num)
                waveform.append(wave)
            print(f'{len(waveform)} waveform data extracted')
            self.filter_data = iter(waveform)
        elif predicate is ismrmrd.Image:
            print(f'{self.dataset.number_of_images(**kwargs)} images included')
            image = []
            for image_num in range(self.dataset.number_of_images(**kwargs)):
                kwargs['imnum'] = image_num
                img = self.dataset.read_image(**kwargs)
                image.append(img)
            print(f'{len(image)} image data extracted')
            self.filter_data = iter(image)
        else:
            print(f'{predicate} unsupported')

    def send(self, image):
        if self.plot_mode:
            print('Plotting images...')
            image_data = image.data.squeeze()

            fig, ax = plt.subplots(1, 1, dpi=150)
            ax.imshow(image_data, 'gray')
            ax.set_aspect(image_data.shape[0] / image_data.shape[1])
            ax.set_title('image')
            ax.axis('off')
            plt.show()
        else:
            print('Saving images...')
            self.save_dataset = ismrmrd.Dataset(self.save_path)
            self.save_dataset.append_image('image', image)

    def close(self):
        self.dataset.close()
        if self.save_dataset:
            self.save_dataset.close()
        print('Dataset closed')

    def __next__(self):
        try:
            return next(self.filter_data)
        except StopIteration:
            print('Iterating data finished by next')
            raise StopIteration("Length too big")

    def __iter__(self):
        while True:
            try:
                yield next(self)
            except StopIteration:
                print('Iterating data finished by iteration')
                return



import argparse
import ismrmrd

import numpy as np
import matplotlib.pyplot as plt


parser = argparse.ArgumentParser()
parser.add_argument('-r', '--raw_path', type=str, default='../../rawdata.h5')

args = parser.parse_args()
fpath = args.raw_path

mrd_data = ismrmrd.Dataset(fpath)
acq_num = mrd_data.number_of_acquisitions()

xml_header = ismrmrd.xsd.CreateFromDocument(mrd_data.read_xml_header())
matrix_size = xml_header.encoding[0].encodedSpace.matrixSize

acquisitions = []
for num in range(acq_num):
    acquisitions.append(mrd_data.read_acquisition(num))

raw_data = np.zeros((matrix_size.x, matrix_size.y, matrix_size.z, acquisitions[0].data.shape[0]), dtype=np.complex64)



for acq in acquisitions:
    raw_data[:, acq.idx.kspace_encode_step_1, acq.idx.kspace_encode_step_2, :] = acq.data.transpose(1, 0)

print(f'RO:\t\t{raw_data.shape[0]}')
print(f'PE:\t\t{raw_data.shape[1]}')
print(f'Slice:\t\t{raw_data.shape[2]}')
print(f'Channel:\t{raw_data.shape[3]}')


fig, axes = plt.subplots(2, 2)
axes[0, 0].imshow(np.abs(raw_data[:, :, 0, 0]).T, cmap='gray')
axes[0, 0].axis('off')
axes[0, 0].set_title('Amplitude')

axes[0, 1].imshow(np.angle(raw_data[:, :, 0, 0]).T, cmap='gray')
axes[0, 1].axis('off')
axes[0, 1].set_title('Phase')

axes[1, 0].imshow(np.real(raw_data[:, :, 0, 0]).T, cmap='gray')
axes[1, 0].axis('off')
axes[1, 0].set_title('Real')

axes[1, 1].imshow(np.imag(raw_data[:, :, 0, 0]).T, cmap='gray')
axes[1, 1].axis('off')
axes[1, 1].set_title('Imag')

plt.show()



import time
import logging
import ismrmrd
import itertools
import numpy as np

from gadgetron.util.cfft import cfftn, cifftn

def remove_oversampling(acquisitions, header):
    print("Start remove_oversampling.")
    encoding_space = header.encoding[0].encodedSpace.matrixSize
    recon_space = header.encoding[0].reconSpace.matrixSize

    if encoding_space.x == recon_space.x:
        return acquisitions

    x0 = (encoding_space.x - recon_space.x) // 2
    x1 = (encoding_space.x - recon_space.x) // 2 + recon_space.x

    def crop_acquisition(acquisition):
        x_space = cifftn(acquisition.data, axes=[1])
        x_space = x_space[:, x0:x1]
        acquisition.resize(number_of_samples=x_space.shape[1], active_channels=x_space.shape[0])
        acquisition.center_sample = recon_space.x // 2
        acquisition.data[:] = cfftn(x_space, axes=[1])

        return acquisition

    print("End remove_oversampling.")
    return map(crop_acquisition, acquisitions)


def accumulate_acquisitions(acquisitions, header):
    print("Start accumulate_acquisitions.")
    accumulated_acquisitions = []
    matrix_size = header.encoding[0].encodedSpace.matrixSize
    print("matrix_size.x=" + str(matrix_size.x))
    print("matrix_size.y=" + str(matrix_size.y))
    print("matrix_size.z=" + str(matrix_size.z))

    def assemble_buffer(acqs):
        print("Start assemble_buffer.")
        number_of_channels = acqs[0].data.shape[0]
        number_of_samples = acqs[0].data.shape[1]

        buffer = np.zeros(
            (number_of_channels,
             number_of_samples,
             matrix_size.y,
             matrix_size.z),
            dtype=np.complex64
        )

        for acq in acqs:
            buffer[:, :, acq.idx.kspace_encode_step_1, acq.idx.kspace_encode_step_2] = acq.data

        print("End assemble_buffer.")
        return buffer

    for acquisition in acquisitions:
        accumulated_acquisitions.append(acquisition)
        if acquisition.is_flag_set(ismrmrd.ACQ_LAST_IN_SLICE):
            yield acquisition, assemble_buffer(accumulated_acquisitions)
            accumulated_acquisitions = []

    print("End accumulate_acquisitions.")

def reconstruct_images(buffers, header):
    print("Start reconstruct_images.")
    indices = itertools.count(start=1)
    field_of_view = header.encoding[0].reconSpace.fieldOfView_mm

    def reconstruct_image(kspace_data):
        return cifftn(kspace_data, axes=[1, 2, 3])

    def combine_channels(image_data):
        # return np.sqrt(np.sum(np.square(np.abs(image_data)), axis=0))
        return image_data[0]
        # return image_data

    for reference, data in buffers:
        print("image_index = " + str(indices))
        yield ismrmrd.image.Image.from_array(
            combine_channels(reconstruct_image(data)),
            acquisition=reference,
            image_index=next(indices),
            image_type=ismrmrd.IMTYPE_MAGNITUDE,
            # data_type=ismrmrd.DATATYPE_FLOAT,
            data_type=ismrmrd.DATATYPE_CXFLOAT,
            field_of_view=(field_of_view.x, field_of_view.y, field_of_view.z),
            transpose = False
        )
    
    print("End reconstruct_images.")


def recon_acquisitions(connection):
    print("Python reconstruction running - reconstructing images from acquisitions.")

    start = time.time()

    connection.filter(ismrmrd.Acquisition)

    acquisitions = iter(connection)
    acquisitions = remove_oversampling(acquisitions, connection.header)
    buffers = accumulate_acquisitions(acquisitions, connection.header)
    images = reconstruct_images(buffers, connection.header)

    for image in images:
        print("Sending image back to client.")
        print(f"image.matrix_size = '{image.matrix_size}'")
        connection.send(image)

    print("Python reconstruction done.")



import faulthandler
import logging
import multiprocessing

# from PIL import Image as PILImage
import gadgetron

def do_work(conn):
    print(rf'iam in')
    for index,acq in enumerate(conn):
        print(rf'{index}: data type: {type(acq)}')
        conn.send(acq)
        pass
    pass

if __name__ == "__main__":
    faulthandler.enable()
    logging.basicConfig(level=logging.INFO, format="[%(levelname)s] %(message)s")

    def spawn_process(*args):
        child = multiprocessing.Process(target=do_work, args=args)
        child.start()

    while True:
        gadgetron.external.listen(18000, spawn_process)
